# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import contextvars
import dataclasses
import functools
import os
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import Optional

import requests

import kolena
import kolena._api.v1.token as API
from kolena._utils import krequests
from kolena._utils.serde import from_dict
from kolena.errors import InvalidClientStateError
from kolena.errors import InvalidTokenError
from kolena.errors import UnauthenticatedError
from kolena.errors import UninitializedError

API_V1 = "v1"
API_V2 = "v2"
DEFAULT_API_VERSION = API_V1
API_URL = "https://api.kolena.io"
API_URL_ENV_VAR = "KOLENA_API_URL"
CLIENT_STATE: contextvars.ContextVar["_ClientState"] = contextvars.ContextVar("client_state")


class NoOpAuth(requests.auth.AuthBase):
    """Provide no-op Auth to disable requests using .netrc"""

    def __call__(self, r: Any) -> Any:
        return r


class _ClientState:
    def __init__(
        self,
        base_url: Optional[str] = API_URL,
        api_token: Optional[str] = None,
        jwt_token: Optional[str] = None,
        tenant: Optional[str] = None,
        verbose: bool = False,
        telemetry: bool = False,
        proxies: Optional[Dict[str, str]] = None,
        additional_request_headers: Optional[Dict[str, Any]] = None,
    ):
        self.base_url: Optional[str] = None
        self.api_token: Optional[str] = None
        self.jwt_token: Optional[str] = None
        self.tenant: Optional[str] = None
        self.verbose: bool = False
        self.telemetry: bool = False
        self.proxies: Dict[str, str] = {}
        self.additional_request_headers: Optional[Dict[str, Any]] = None
        self.update(
            base_url=base_url,
            api_token=api_token,
            jwt_token=jwt_token,
            tenant=tenant,
            verbose=verbose,
            telemetry=telemetry,
            proxies=proxies,
            additional_request_headers=additional_request_headers,
        )

    def update(
        self,
        base_url: Optional[str] = None,
        api_token: Optional[str] = None,
        jwt_token: Optional[str] = None,
        tenant: Optional[str] = None,
        verbose: bool = False,
        telemetry: bool = False,
        proxies: Optional[Dict[str, str]] = None,
        additional_request_headers: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.base_url = base_url or self.base_url
        self.api_token = api_token or self.api_token
        self.jwt_token = jwt_token or self.jwt_token
        self.tenant = tenant or self.tenant
        self.verbose = verbose
        self.telemetry = telemetry
        self.proxies = proxies or self.proxies
        self.additional_request_headers = additional_request_headers or self.additional_request_headers

    def assert_initialized(self) -> None:
        if self.base_url is None:
            raise InvalidClientStateError("missing base_url")
        if self.jwt_token is None:
            raise UninitializedError("client has not been initialized via kolena.initialize(...)")
        if self.api_token is None:
            raise InvalidClientStateError("missing client api_token")

    def reset(self) -> None:
        # note that base_url remains set
        self.api_token = None
        self.jwt_token = None
        self.tenant = None
        self.verbose = False
        self.telemetry = False
        self.additional_request_headers = None
        self.proxies = {}


def _get_api_base_url() -> str:
    return os.environ.get(API_URL_ENV_VAR) or API_URL


_client_state = _ClientState(base_url=_get_api_base_url())


def get_client_state() -> _ClientState:
    return CLIENT_STATE.get(_client_state)


def is_client_initialized() -> bool:
    try:
        get_client_state().assert_initialized()
    except (InvalidClientStateError, UninitializedError):
        return False
    return True


def kolena_initialized(func: Callable) -> Callable:
    """Raises InvalidKolenaStateError if not initialized"""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        client_state = get_client_state()
        client_state.assert_initialized()
        return func(*args, **kwargs)

    return wrapper


def get_endpoint_with_baseurl(base_url: str, endpoint_path: str, api_version: str = DEFAULT_API_VERSION) -> str:
    return f"{base_url}/{api_version}/{endpoint_path.lstrip('/')}"


def get_token(
    api_token: str,
    base_url: Optional[str] = None,
    proxies: Optional[Dict[str, str]] = None,
) -> API.ValidateResponse:
    base_url = base_url or get_client_state().base_url or ""
    request = API.ValidateRequest(api_token=api_token, version=kolena.__version__)
    r = requests.put(
        get_endpoint_with_baseurl(base_url, "token/login"),
        auth=NoOpAuth(),
        json=dataclasses.asdict(request),
        proxies=proxies,
    )
    try:
        krequests.raise_for_status(r)
    except UnauthenticatedError as e:
        raise InvalidTokenError(e)

    init_response = from_dict(data_class=API.ValidateResponse, data=r.json())
    return init_response


@contextlib.contextmanager
def kolena_session(
    api_token: str,
    base_url: Optional[str] = None,
    additional_request_headers: Optional[Dict[str, Any]] = None,
    proxies: Optional[Dict[str, str]] = None,
) -> Iterator[_ClientState]:
    base_url = base_url or _get_api_base_url()
    init_response = get_token(api_token, base_url)
    client_state = _ClientState(
        base_url=base_url,
        api_token=api_token,
        jwt_token=init_response.access_token,
        tenant=init_response.tenant,
        additional_request_headers=additional_request_headers,
        proxies=proxies,
    )
    token = CLIENT_STATE.set(client_state)

    try:
        yield client_state
    finally:
        CLIENT_STATE.reset(token)



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Reference for various exceptions raised from `kolena`. All custom exceptions extend the base
[`KolenaError`][kolena.errors.KolenaError].
"""
from requests import HTTPError


class KolenaError(Exception):
    """Base error for all Kolena errors to extend. Allows consumers to catch Kolena specific errors."""


class InputValidationError(ValueError, KolenaError):
    """Exception indicating that provided input data failed validation."""


class IncorrectUsageError(RuntimeError, KolenaError):
    """Exception indicating that the user performed a disallowed action with the client."""


class InvalidTokenError(ValueError, KolenaError):
    """Exception indicating that provided token value was invalid."""


class InvalidClientStateError(RuntimeError, KolenaError):
    """Exception indicating that client state was invalid."""


class MissingTokenError(KeyError, KolenaError):
    """Exception indicating that the client could not locate an API token."""


class UninitializedError(InvalidClientStateError):
    """Exception indicating that the client has not been properly initialized before usage."""


class DirectInstantiationError(RuntimeError, KolenaError):
    """
    Exception indicating that the default constructor was used for a class that does not support direct instantiation.
    Available static constructors should be used when this exception is encountered.
    """


class FrozenObjectError(RuntimeError, KolenaError):
    """Exception indicating that the user attempted to modify a frozen object."""


class UnauthenticatedError(HTTPError, KolenaError):
    """Exception indicating unauthenticated usage of the client."""


class RemoteError(HTTPError, KolenaError):
    """Exception indicating that a remote error occurred in communications between the Kolena client and server."""


class CustomMetricsException(KolenaError):
    """Exception indicating that there's an error when computing custom metrics."""


class WorkflowMismatchError(KolenaError):
    """Exception indicating a workflow mismatch."""


class NotFoundError(RemoteError):
    """Exception indicating an entity is not found"""


class NameConflictError(RemoteError):
    """Exception indicating the name of an entity is conflict"""



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from dataclasses import dataclass
from dataclasses import fields
from typing import Any

from kolena._utils.datatypes import _register_data_type
from kolena._utils.datatypes import DataType
from kolena._utils.datatypes import TypedDataObject


class PreventThresholdOverrideMeta(ABCMeta, type):
    def __new__(cls, name: str, bases: tuple, dct: dict) -> "PreventThresholdOverrideMeta":
        if "threshold" in dct.get("__annotations__", {}):
            for base in bases:
                if base.__name__ == "ThresholdedMetrics":
                    raise TypeError(f"Subclasses of {base.__name__} cannot override 'threshold'")
        return super().__new__(cls, name, bases, dct)


class _MetricsType(DataType):
    THRESHOLDED = "THRESHOLDED"

    @staticmethod
    def _data_category() -> str:
        return "METRICS"


@dataclass(frozen=True)
class ThresholdedMetrics(TypedDataObject[_MetricsType], metaclass=PreventThresholdOverrideMeta):
    """
    Represents metrics tied to a specific threshold.

    `List[ThresholdedMetrics]` should be used as a field type within `MetricsTestSample`
    from the `kolena.workflow` module. This list is meant to hold metric values
    associated with distinct thresholds. These metrics are expected to be uniform across `TestSample`
    instances within a single test execution.

    `ThresholdedMetrics` prohibits the use of dictionary objects as field values and guarantees that
    the threshold values remain immutable once set. For application within a particular workflow,
    subclassing is required to define relevant metrics fields.

    Usage example:

    ```python
    from kolena.workflow import MetricsTestSample
    from kolena._experimental.workflow import ThresholdedMetrics

    @dataclass(frozen=True)
    class ClassThresholdedMetrics(ThresholdedMetrics):
        precision: float
        recall: float
        f1: float

    @dataclass(frozen=True)
    class TestSampleMetrics(MetricsTestSample):
        car: List[ClassThresholdedMetrics]
        pedestrian: List[ClassThresholdedMetrics]

    # Creating an instance of metrics
    metric = TestSampleMetrics(
        car=[
            ClassThresholdedMetrics(threshold=0.3, precision=0.5, recall=0.8, f1=0.615),
            ClassThresholdedMetrics(threshold=0.4, precision=0.6, recall=0.6, f1=0.6),
            ClassThresholdedMetrics(threshold=0.5, precision=0.8, recall=0.4, f1=0.533),
            # ...
        ],
        pedestrian=[
            ClassThresholdedMetrics(threshold=0.3, precision=0.6, recall=0.9, f1=0.72),
            ClassThresholdedMetrics(threshold=0.4, precision=0.7, recall=0.7, f1=0.7),
            ClassThresholdedMetrics(threshold=0.5, precision=0.8, recall=0.6, f1=0.686),
            # ...
        ],
    )
    ```

    Raises:
        TypeError: If any of the field values is a dictionary.
    """

    threshold: float

    def __init_subclass__(cls, **kwargs: Any):
        _register_data_type(cls)

    @classmethod
    def _data_type(cls) -> _MetricsType:
        """
        Returns the type of metrics represented by this class.

        Returns:
            _MetricsType: An enumeration value indicating the type of metrics.
        """
        return _MetricsType.THRESHOLDED

    def __post_init__(self) -> None:
        for field in fields(self):
            field_value = getattr(self, field.name)
            if isinstance(field_value, dict):
                raise TypeError(f"Field '{field.name}' should not be a dictionary")



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import mimetypes
import sys
from dataclasses import asdict
from enum import Enum
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from urllib.parse import urlparse

import pandas as pd
import requests

from kolena._api.v1.event import EventAPI
from kolena._api.v2.dataset import CommitData
from kolena._api.v2.dataset import EntityData
from kolena._api.v2.dataset import ListCommitHistoryRequest
from kolena._api.v2.dataset import ListCommitHistoryResponse
from kolena._api.v2.dataset import LoadDatapointsRequest
from kolena._api.v2.dataset import LoadDatasetByNameRequest
from kolena._api.v2.dataset import Path
from kolena._api.v2.dataset import RegisterRequest
from kolena._utils import krequests_v2 as krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.datatypes import _deserialize_dataobject
from kolena._utils.datatypes import _serialize_dataobject
from kolena._utils.datatypes import DATA_TYPE_FIELD
from kolena._utils.datatypes import TypedDataObject
from kolena._utils.endpoints import get_dataset_url
from kolena._utils.instrumentation import with_event
from kolena._utils.serde import from_dict
from kolena._utils.state import API_V2
from kolena.dataset._common import COL_DATAPOINT
from kolena.dataset._common import COL_DATAPOINT_ID_OBJECT
from kolena.dataset._common import DEFAULT_SOURCES
from kolena.dataset._common import validate_batch_size
from kolena.dataset._common import validate_dataframe_ids
from kolena.errors import InputValidationError
from kolena.io import _dataframe_object_serde

_FIELD_LOCATOR = "locator"
_FIELD_TEXT = "text"
_SEP = "."


class DatapointType(str, Enum):
    AUDIO = "DATAPOINT/AUDIO"
    COMPOSITE = "DATAPOINT/COMPOSITE"
    DOCUMENT = "DATAPOINT/DOCUMENT"
    IMAGE = "DATAPOINT/IMAGE"
    POINT_CLOUD = "DATAPOINT/POINT_CLOUD"
    TABULAR = "DATAPOINT/TABULAR"
    TEXT = "DATAPOINT/TEXT"
    VIDEO = "DATAPOINT/VIDEO"

    @classmethod
    def has_value(cls, item: Any) -> bool:
        return item in cls.__members__.values()


_DATAPOINT_TYPE_MAP = {
    "image": DatapointType.IMAGE.value,
    "application/pdf": DatapointType.DOCUMENT.value,
    "text": DatapointType.DOCUMENT.value,
    "video": DatapointType.VIDEO.value,
    "audio": DatapointType.AUDIO.value,
}


def _dataobject_type(obj: TypedDataObject) -> str:
    obj_type = obj._data_type()
    return f"{obj_type._data_category()}/{obj_type.value}"


def _get_datapoint_type(mimetype_str: str) -> Optional[str]:
    main_type, sub_type = mimetype_str.split("/")
    return _DATAPOINT_TYPE_MAP.get(mimetype_str, None) or _DATAPOINT_TYPE_MAP.get(main_type, None)


def _normalize_url(x: str) -> str:
    url = urlparse(x)
    return url._replace(query="", fragment="").geturl()


def _infer_datatype_value(x: str) -> str:
    url = _normalize_url(x)
    mtype, _ = mimetypes.guess_type(url)
    if mtype:
        datatype = _get_datapoint_type(mtype)
        if datatype is not None:
            return datatype
    elif url.endswith(".pcd"):
        return DatapointType.POINT_CLOUD.value

    return DatapointType.TABULAR.value


def _add_datatype(df: pd.DataFrame) -> None:
    """Adds `data_type` column(s) to input DataFrame."""
    prefixes = {
        column.rsplit(sep=_SEP, maxsplit=1)[0]
        for column in df.columns.values
        if isinstance(column, str) and _SEP in column
    }
    if prefixes:
        df[DATA_TYPE_FIELD] = DatapointType.COMPOSITE.value
        for prefix in prefixes:
            if not prefix.strip():
                raise InputValidationError(
                    "Empty prefix encountered when parsing composite dataset. "
                    f"Columns must lead with at least one non-whitespace character prior to delimeter '{_SEP}'.",
                )
            if prefix in df.columns:
                raise InputValidationError(
                    f"Conflicting column '{prefix}' encountered when formatting composite dataset.",
                )
            if _SEP in prefix:
                raise InputValidationError(
                    f"More than one delimeter '{_SEP}' in prefix: '{prefix}'.",
                )

            composite_columns = df.filter(regex=rf"^{prefix}", axis=1).columns.to_list()
            composite = df.loc[:, composite_columns].rename(columns=lambda col: col.split(_SEP)[-1])
            composite[DATA_TYPE_FIELD] = _infer_datatype(composite)

            df[prefix] = composite.to_dict("records")
            df.drop(columns=composite_columns, inplace=True)
    else:
        df[DATA_TYPE_FIELD] = _infer_datatype(df)


def _infer_datatype(df: pd.DataFrame) -> Union[pd.DataFrame, str]:
    if _FIELD_LOCATOR in df.columns:
        return df[_FIELD_LOCATOR].apply(_infer_datatype_value)
    elif _FIELD_TEXT in df.columns:
        return DatapointType.TEXT.value

    return DatapointType.TABULAR.value


def _infer_id_fields(df: pd.DataFrame) -> List[str]:
    def get_id_fields_by(field: str) -> List[str]:
        return [
            id_field
            for id_field in df.columns.array
            if isinstance(id_field, str) and id_field.rsplit(_SEP, maxsplit=1)[-1] == field
        ]

    if id_fields := get_id_fields_by(_FIELD_LOCATOR):
        return id_fields
    elif id_fields := get_id_fields_by(_FIELD_TEXT):
        return id_fields
    raise InputValidationError("Failed to infer the id_fields, please provide id_fields explicitly")


def _to_serialized_dataframe(df: pd.DataFrame, column: str) -> pd.DataFrame:
    result = _dataframe_object_serde(df, _serialize_dataobject)
    if column == COL_DATAPOINT:
        _add_datatype(result)
    result[column] = result.to_dict("records")
    result[column] = result[column].apply(lambda x: json.dumps(x))

    return result[[column]]


def _to_deserialized_dataframe(df: pd.DataFrame, column: str) -> pd.DataFrame:
    flattened = pd.json_normalize(
        [json.loads(r[column]) if r[column] is not None else {} for r in df.to_dict("records")],
        max_level=0,
    )
    flattened = _flatten_composite(flattened)
    flattened = flattened.loc[:, ~flattened.columns.str.endswith(DATA_TYPE_FIELD)]
    return _dataframe_object_serde(flattened, _deserialize_dataobject)


def _flatten_composite(df: pd.DataFrame) -> pd.DataFrame:
    for key, value in df.iloc[0].items():
        if isinstance(value, dict) and DatapointType.has_value(value.get(DATA_TYPE_FIELD)):
            flattened = pd.json_normalize(df[key], max_level=0).rename(
                columns=lambda col: f"{key}{_SEP}{col}",
            )
            df = df.join(flattened)
            df.drop(columns=[key], inplace=True)
    return df


def _upload_dataset_chunk(df: pd.DataFrame, load_uuid: str, id_fields: List[str]) -> None:
    df_serialized_datapoint = _to_serialized_dataframe(df, column=COL_DATAPOINT)
    df_serialized_datapoint_id_object = _to_serialized_dataframe(df[sorted(id_fields)], column=COL_DATAPOINT_ID_OBJECT)
    df_serialized = pd.concat([df_serialized_datapoint, df_serialized_datapoint_id_object], axis=1)

    upload_data_frame(df=df_serialized, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=load_uuid)


def _load_dataset_metadata(name: str) -> Optional[EntityData]:
    """
    Load the metadata of a given dataset.

    :param name: The name of the dataset.
    :return: The metadata of the dataset.
    """
    response = krequests.put(Path.LOAD_DATASET, json=asdict(LoadDatasetByNameRequest(name=name)))
    if response.status_code == requests.codes.not_found:
        return None

    response.raise_for_status()

    return from_dict(EntityData, response.json())


def _resolve_id_fields(
    df: pd.DataFrame,
    id_fields: Optional[List[str]],
    existing_dataset: Optional[EntityData],
) -> List[str]:
    existing_id_fields = []
    if existing_dataset:
        existing_id_fields = existing_dataset.id_fields
    if not id_fields:
        if existing_id_fields:
            return existing_id_fields
        else:
            id_fields = _infer_id_fields(df)
    return id_fields


def _prepare_upload_dataset_request(
    name: str,
    df: Union[pd.DataFrame, Iterator[pd.DataFrame]],
    *,
    id_fields: Optional[List[str]] = None,
) -> Tuple[List[str], str]:
    load_uuid = init_upload().uuid
    existing_dataset = _load_dataset_metadata(name)
    if isinstance(df, pd.DataFrame):
        id_fields = _resolve_id_fields(df, id_fields, existing_dataset)
        validate_dataframe_ids(df, id_fields)
        _upload_dataset_chunk(df, load_uuid, id_fields)
    else:
        validated = False
        for chunk in df:
            if not validated:
                id_fields = _resolve_id_fields(chunk, id_fields, existing_dataset)
                validate_dataframe_ids(chunk, id_fields)
                validated = True
            assert id_fields is not None
            _upload_dataset_chunk(chunk, load_uuid, id_fields)
    assert id_fields is not None
    return id_fields, load_uuid


def _send_upload_dataset_request(
    name: str,
    id_fields: List[str],
    load_uuid: str,
    sources: Optional[List[Dict[str, str]]],
) -> EntityData:
    request = RegisterRequest(name=name, id_fields=id_fields, uuid=load_uuid, sources=sources)
    response = krequests.post(Path.REGISTER, json=asdict(request))
    krequests.raise_for_status(response)
    data = from_dict(EntityData, response.json())
    return data


def _upload_dataset(
    name: str,
    df: Union[pd.DataFrame, Iterator[pd.DataFrame]],
    *,
    id_fields: Optional[List[str]] = None,
    sources: Optional[List[Dict[str, str]]] = DEFAULT_SOURCES,
) -> None:
    prepared_id_fields, load_uuid = _prepare_upload_dataset_request(name, df, id_fields=id_fields)

    data = _send_upload_dataset_request(name, prepared_id_fields, load_uuid, sources=sources)
    log.info(f"uploaded dataset '{name}' ({get_dataset_url(dataset_id=data.id)})")


@with_event(event_name=EventAPI.Event.REGISTER_DATASET)
def upload_dataset(
    name: str,
    df: Union[pd.DataFrame, Iterator[pd.DataFrame]],
    *,
    id_fields: Optional[List[str]] = None,
) -> None:
    """
    Create or update a dataset with the contents of the provided DataFrame `df`.

    !!! note "Updating `id_fields`"
        ID fields are used to associate model results (uploaded via [`upload_results`][kolena.dataset.upload_results])
        with datapoints in this dataset. When updating an existing dataset, update `id_fields` with caution.

    :param name: The name of the dataset.
    :param df: A DataFrame or iterator of DataFrames. Provide an iterator to perform batch upload (example:
        `csv_reader = pd.read_csv("PathToDataset.csv", chunksize=10)`).
    :param id_fields: Optionally specify a list of ID fields that will be used to link model results with the datapoints
        within a dataset. When unspecified, a suitable value is inferred from the columns of the provided `df`. Note
        that `id_fields` must be hashable.
    """
    _upload_dataset(name, df, id_fields=id_fields)


def _iter_dataset_raw(
    name: str,
    commit: Optional[str] = None,
    batch_size: int = BatchSize.LOAD_SAMPLES.value,
) -> Iterator[pd.DataFrame]:
    validate_batch_size(batch_size)
    init_request = LoadDatapointsRequest(
        name=name,
        commit=commit,
        batch_size=batch_size,
    )
    yield from _BatchedLoader.iter_data(
        init_request=init_request,
        endpoint_path=Path.LOAD_DATAPOINTS.value,
        df_class=None,
        endpoint_api_version=API_V2,
    )


def _iter_dataset(
    name: str,
    commit: Optional[str] = None,
    batch_size: int = BatchSize.LOAD_SAMPLES.value,
) -> Iterator[pd.DataFrame]:
    """
    Get an iterator over datapoints in the dataset.
    """
    for df_batch in _iter_dataset_raw(name, commit, batch_size):
        yield _to_deserialized_dataframe(df_batch, column=COL_DATAPOINT)


@with_event(event_name=EventAPI.Event.FETCH_DATASET)
def download_dataset(name: str, *, commit: Optional[str] = None) -> pd.DataFrame:
    """
    Download an entire dataset given its name.

    :param name: The name of the dataset.
    :param commit: The commit hash for version control. Get the latest commit when this value is `None`.
    :return: A DataFrame containing the specified dataset.
    """
    df_batches = list(_iter_dataset(name, commit, BatchSize.LOAD_SAMPLES.value))
    log.info(f"downloaded dataset '{name}'")
    return pd.concat(df_batches, ignore_index=True) if df_batches else pd.DataFrame()


def _list_commits(name: str, descending: bool = False, offset: int = 0, limit: int = 50) -> ListCommitHistoryResponse:
    """
    Invoke the list-commits api.
    """
    request = ListCommitHistoryRequest(name=name, descending=descending, offset=offset, limit=limit)
    response = krequests.put(Path.LIST_COMMITS, json=asdict(request))
    krequests.raise_for_status(response)
    return from_dict(ListCommitHistoryResponse, response.json())


def _iter_commits(
    name: str,
    descending: bool = False,
    limit: Optional[int] = None,
    page_size: int = 50,
) -> Iterator[Tuple[int, List[CommitData]]]:
    """
    Get an iterator over the commit history of the dataset.
    """
    current_count = 0
    if limit is None:
        limit = sys.maxsize
    while True:
        response = _list_commits(name, descending=descending, offset=current_count, limit=page_size)
        yield response.total_count, response.records
        current_count += len(response.records)
        if current_count >= min(limit, response.total_count):
            break


@with_event(event_name=EventAPI.Event.FETCH_DATASET_HISTORY)
def _fetch_dataset_history(
    name: str,
    *,
    descending: bool = False,
    limit: Optional[int] = None,
    page_size: int = 50,
) -> Tuple[int, List[CommitData]]:
    """
    Get the commit history of a dataset.

    :param name: The name of the dataset
    :param descending: If True, return the results in descending order.
    :param limit: The maximum number of results to return. If None, all results will be returned.
    :param page_size: The number of items to return per page.
    :return: A tuple where the first element is the total number of commits
             and the second element is a list of CommitData objects.
    """
    iter_commit_responses = list(_iter_commits(name, descending, limit, page_size))
    total_commit_count = iter_commit_responses[0][0]
    commits = [commit for response in iter_commit_responses for commit in response[1]][:limit]
    return total_commit_count, commits
